'''
Created on 23 Jan 2013

@author: gabs
'''

import numpy as np
import sys
from itertools import izip, imap
from scipy import optimize
import bayesian_track_analysis as bta

PI = np.pi

# FUNCTION: optimisetransitionparameters

def optimisetransitionparameters(empirical_priors, tracks_rt):
    
    # define function to minimize
    hmm_fn = lambda x: HmmFun(x[0], x[1], empirical_priors, tracks_rt, True)
    
    # initial guess: find approximate minimum
    x,y = np.meshgrid(np.linspace(0.001,0.999,10),np.linspace(0.001,0.999,10))
    z = Hmm(empirical_priors, tracks_rt, (x,y) )
    min_idx = np.argmin(z)
    
    x0 = np.array([x.flat[min_idx], y.flat[min_idx]])
    
    imode = 8
    eps = 2e-8
    num_iter=0
    while imode>7:
        eps = eps/2
        res, fx, its, imode, smode = \
        optimize.fmin_slsqp(hmm_fn, x0, eqcons=[], f_eqcons=None, ieqcons=[], f_ieqcons=None, bounds=[(0., 1.), (0., 1.)], \
                              fprime=None, fprime_eqcons=None, fprime_ieqcons=None, args=(), iter=1000, acc=1e-06, iprint=1, \
#                              disp=None, \
full_output=True, epsilon=eps )
        num_iter += 1
        assert num_iter<100, 'Optimisation failed'
            
      

    return res
        
# END FUNCTION

# FUNCTION: optimisetransitionparameterswithbootstrap
# gives an estimate of the CI of the parameters

def optimisetransitionparameterswithbootstrap(empirical_priors, tracks_rt, bootstrap_B=999):
    
    bs_params = []
    n = len(tracks_rt)
    
    # run on real data
    true_params = optimisetransitionparameters(empirical_priors, tracks_rt)    
    count = 0
    
    while count<bootstrap_B:
        bs_idx = np.random.randint(0, n, n)
        bs_tracks = [tracks_rt[x] for x in bs_idx]
        hmm_fn = lambda x: HmmFun(x[0], x[1], empirical_priors, bs_tracks, True)
        res, fx, its, imode, smode = \
        optimize.fmin_slsqp(hmm_fn, true_params, bounds=[(0., 1.), (0., 1.)], iter=1000, full_output=True)
        if imode==0:
            bs_params.append( res )
            count += 1
    
    return bs_params, true_params
    
# END FUNCTION

# FUNCTION: HMM interface
def Hmm(empirical_priors, tracks_rt, probs, b_likelihood_only=True):
    
    # unpack probs
    p01, p10 = probs
    
    # check type
    if not(isinstance(p01, float)):
        # NOT of type float
        assert isinstance(p01, np.ndarray), 'multiple transition probabilities must be supplied in a numpy array'
        if not(b_likelihood_only):
            # can't handle this case
            print 'multiple transition probabilities supplied and b_likelihood_only=False'
            raise Exception()
        shape01 = p01.shape
        shape10 = p10.shape    
        assert shape01==shape10, 'dimensions of P01 do not match those of P10'
        b_single_prob_pair = False
    else:
        # type float
        b_single_prob_pair = True
    
    if b_likelihood_only:
        if b_single_prob_pair:
            neg_log_likelihood = HmmFun(p01, p10, empirical_priors, tracks_rt, b_likelihood_only)
        else:
            num_prob = p01.size
            
            # testing iterators for array creation
            hmm_fn = lambda x, y: HmmFun(x, y, empirical_priors, tracks_rt, b_likelihood_only)
            neg_log_likelihood = np.array([hmm_fn(i,j) for i,j in izip(p01.flat, p10.flat)])
            
            # reshape nll
            neg_log_likelihood = neg_log_likelihood.reshape(p01.shape)
            return neg_log_likelihood
                
    else: # NOT b_likelihood_only
        # to get to here, probs must be floats
        run_probs, neg_log_likelihood = HmmFun(p01, p10, empirical_priors, tracks_rt, b_likelihood_only)
        return run_probs, neg_log_likelihood
    
    # END IF B_LIKELIHOOD_ONLY
        
# END FUNCTION

# FUNCTION: HMM CPP interface to iterate over tracks
def HmmIterateTracksCpp(p01, p10, empirical_priors, tracks_rt, b_likelihood_only=True):
    
    # unpack empirical_priors tuple
    displacement_pdf0, angle_pdf0, \
    displacement_pdf1, angle_pdf1 = empirical_priors
    
    # convert to Cpp linear interpolant object
    displacement_map_nm = hmm_cpp.LinInterp();
    displacement_map_nc = hmm_cpp.LinInterp();
    angle_map_nm = hmm_cpp.LinInterp();
    angle_map_nc = hmm_cpp.LinInterp();
    
    for i in range(np.size(displacement_pdf0.x)):
        displacement_map_nm[displacement_pdf0.x[i]] = displacement_pdf0.y[i]
        
    for i in range(np.size(displacement_pdf1.x)):
        displacement_map_nc[displacement_pdf1.x[i]] = displacement_pdf1.y[i]
    
    for i in range(np.size(angle_pdf0.x)):
        angle_map_nm[angle_pdf0.x[i]] = angle_pdf0.y[i]
        
    for i in range(np.size(angle_pdf1.x)):
        angle_map_nc[angle_pdf1.x[i]] = angle_pdf1.y[i]
        
    nll = 0.
    run_probs = []
        
    # LOOP over tracks
    
    for this_track_rt in tracks_rt:
    
        this_track_r = hmm_cpp.VectorOfDouble()
        this_track_t = hmm_cpp.VectorOfDouble()
        this_track_r.extend(this_track_rt[0])
        this_track_t.extend(this_track_rt[1])
    
        if b_likelihood_only:
            nll += hmm_cpp.hmm_main(this_track_r, this_track_t, p01, p10, displacement_map_nm, 
                                displacement_map_nc, angle_map_nm, angle_map_nc, None)
            
            
        else:
            # vector to hold run probs
            this_run_probs = hmm_cpp.VectorOfDouble();
            this_run_probs.extend(np.zeros(np.size(this_track_r)))
            
            nll += hmm_cpp.hmm_main(this_track_r, this_track_t, p01, p10, displacement_map_nm, 
                                    displacement_map_nc, angle_map_nm, angle_map_nc, this_run_probs)
            
            # convert run probs back to np array and store
            run_probs.append(np.array(this_run_probs))
                    
        # END LOOP over tracks
    
    if b_likelihood_only:
        return nll
    else:
        return run_probs, nll
    
# END FUNCTION

# FUNCTION loop over a list of tracks and return run probs (opt) and neg log likelihood
def HmmIterateTracks(p01, p10, empirical_priors, tracks_rt, b_likelihood_only=True):
    if b_likelihood_only:
        return sum([HmmMain(p01, p10, empirical_priors, track, True) for track in tracks_rt])
    else:
        temp = [HmmMain(p01, p10, empirical_priors, track, False) for track in tracks_rt]
        run_probs, nll = zip(*temp)
        run_probs = list(run_probs)
        nll = sum(nll)
        return run_probs, nll
    
# END FUNCTION    

# FUNCTION main HMM code body
def HmmMain(p01, p10, empirical_priors, track_rt, b_likelihood_only):
        
    # for convenience
    p00 = 1-p01
    p11 = 1-p10
    
    # numerical minimum
    eps = 1e-12
    
    # unpack empirical_priors tuple
    displacement_pdf0, angle_pdf0, \
    displacement_pdf1, angle_pdf1 = empirical_priors
    
    # initialise key variables
    log_renorm_const = 0.;
    n = track_rt[0,:].size
    alpha0 = np.zeros(n)
    alpha1 = np.zeros(n)
    if not(b_likelihood_only):
        beta0 = np.zeros(n)
        beta1 = np.zeros(n)
        
    # pre-process observation functions

    D0 = displacement_pdf0(track_rt[0,:])
    D1 = displacement_pdf1(track_rt[0,:])
    A0 = angle_pdf0(track_rt[1,:])
    A1 = angle_pdf1(track_rt[1,:])
    
    b00 = D0*A0
    b01 = D1*A0
    b11 = D1*A1
    
    # ignore first entry
    b00[0] = 0.
    b10 = b00
    b01[0] = 0.
    b11[0] = 0.
    
    # initial entry to alpha is special case: use ONLY displacement info
    r = track_rt[0,0]
    alpha0[0] = displacement_pdf0(r)
    alpha1[0] = displacement_pdf1(r)
    
    # first entry to beta is by convention
    if not(b_likelihood_only):
        beta0[n-1] = 1.
        beta1[n-1] = 1.
        
    # loop over all other transitions
    for i in range(1,n):
        
        r = track_rt[0,i]
        t = track_rt[1,i]
        
        alpha0[i] = alpha0[i-1]*p00*b00[i] + alpha1[i-1]*p10*b10[i]
        alpha1[i] = alpha0[i-1]*p01*b01[i] + alpha1[i-1]*p11*b11[i]
        
        norm_const = alpha0[i]+alpha1[i]
        alpha0[i] = alpha0[i]/norm_const
        alpha1[i] = alpha1[i]/norm_const
        log_renorm_const += np.log(norm_const)
        
        if not(b_likelihood_only):

            beta0[n-i-1]=beta0[n-i]*p00*b00[n-i] + beta1[n-i]*p01*b01[n-i]
            beta1[n-i-1]=beta0[n-i]*p10*b10[n-i] + beta1[n-i]*p11*b11[n-i]
            
            norm_const = beta0[n-i-1]+beta1[n-i-1]
            beta0[n-i-1] = beta0[n-i-1]/norm_const
            beta1[n-i-1] = beta1[n-i-1]/norm_const
    
    # END MAIN LOOP
    
    # return variables
    if b_likelihood_only:
        return -log_renorm_const
    else:
#        for i in range(n):
#            print "alpha0(", i, ") =", alpha0[i]
#            print "alpha1(", i, ") =", alpha1[i]
#            print "beta0(", i, ") =", beta0[i]
#            print "beta1(", i, ") =", beta1[i]
        A = alpha1*beta1
        B = A + alpha0*beta0
        run_probs = A/B
        return run_probs, -log_renorm_const

# END FUNCTION

if bta.b_use_ext_module:
    import hmm_cpp
    HmmFun = HmmIterateTracksCpp
else:
    HmmFun = HmmIterateTracks
