#!/usr/bin/env python

'''
Created on 27 Feb 2013

@author: Gabriel Rosser
@email: gabriel.rosser@gmail.com

Contains usage documentation and examples for analysis code
'''

#import bayesian_track_analysis as bta

from bayesian_track_analysis import io, process, plot, hmm
from scipy.interpolate import interp1d
import matplotlib.pyplot as plt
import numpy as np
import csv

############################################
### STAGE 1: compute priors
############################################

# paths to CSV files for nonmotile (1) and nonchemotactic (2) mutants
# CSV files are in the following form:
# track(1).x(1), track(1).x(2), ... , track(1).x(end)
# track(1).y(1), track(1).y(2), ... , track(1).y(end)
# track(2).x(1), track(2).x(2), ... , track(2).x(end)
# track(2).y(1), track(2).y(2), ... , track(2).y(end)
# etc.  

# load tracks

# from CSV

# csv_filename_nm = "data/nm.csv"
# csv_filename_nc = "data/nc.csv"
# csv_filename_wt = "data/wt.csv"
# tracks_nm = io.loadtrackscsv(csv_filename_nm)
# tracks_nc = io.loadtrackscsv(csv_filename_nc)
# tracks_wt = io.loadtrackscsv(csv_filename_wt)

# from Matlab .mat file

mat_filename = "data/short_rhodobacter_tracks.mat"
tracks_nm = io.loadtracksmat(mat_filename, 'Tracks_nm') # first variable is the .mat filename, second is the name of the variable within that file
tracks_nc = io.loadtracksmat(mat_filename, 'Tracks_nc')
tracks_wt = io.loadtracksmat(mat_filename, 'Tracks_wt')

# compute empirical priors (observation pdfs)

empirical_priors = process.computepriors(tracks_nm, tracks_nc, tracks_wt)

# if using the  speed-only method, replace the run_angles_pdf and stop_angles_pdf in empirical_priors

if 0:
    speed_only_angle_pdf = interp1d(np.pi*np.array([-1, 1]), 1/(2*np.pi)*np.ones(2), kind='zero')
    empirical_priors = empirical_priors[0], speed_only_angle_pdf, empirical_priors[2], speed_only_angle_pdf 

# optionally plot the priors
# all plots are displayed when plt.show() is called at the end

plot.plotpriors(empirical_priors)

############################################
### STAGE 2: Optimise transition probs and segment tracks
############################################

# tracks_wt is a list containing the unmodified tracks loaded from CSV
# tracks_wt_rt is the modified polar representation
tracks_wt_rt = process.trackcarttopolar(tracks_wt)

# optimise the transition probabilities

transition_probs = hmm.optimisetransitionparameters(empirical_priors, tracks_wt_rt)

# NB if an estimate of the error in optimisation is required, it is possilbe to use a bootstrap method:

# bs_transition_probs, transition_probs = hmm.optimisetransitionparameterswithbootstrap(empirical_priors, tracks_wt_rt)

# use transition probabilities to segment tracks

run_probs, negative_log_likelihood = hmm.Hmm(empirical_priors, tracks_wt_rt, transition_probs, False)

# output run probabilities to CSV

run_probs_out_file = 'run_probs.csv'
io.saverunprobstocsv(run_probs, run_probs_out_file, b_overwrite=True)

# plot 20 of the tracks in colour to show the run probabilities

plot.plotsometrackstateprobs(tracks_wt, run_probs, 20)
    
############################################
### STAGE 3: Post-processing
############################################

# compute various properties relating to the inferred running and stopping phases
# each variable is an array containing all observed values

run_displacements, stop_displacements, run_angles, stop_angles = process.computesegmenteddispandangle(tracks_wt, run_probs)
run_durations, stop_durations = process.computesegmentdurations(tracks_wt, run_probs)

# compute statistics relating specifically to run-stop-run events (reorientation events)
# each variable is a list of arrays, with each array corresponding to a track and each element in the array corresponding
# a single reorientation event in that track.

stopidx, entryfwdisplacement, exitfwdisplacement, swanglechange, swduration = process.computestopwisequantities(tracks_wt, run_probs)

# for many purposes, we just require a concatenated list of the data, achieved like this

sw_angles_cat = []
[sw_angles_cat.extend(x) for x in swanglechange]

# output stopwise angle changes to CSV

sw_angle_out_file = 'sw_angle.csv'
with open(sw_angle_out_file, 'w') as f:
    csv.writer(f).writerow(sw_angles_cat)
    f.close()

# plot histogram
plt.figure()
plt.hist(sw_angles_cat, 50)

# render all plots

plt.show()